; ---------------------------------------------------------------
; The beginnings of a universal function library
; ---------------------------------------------------------------
ORG 		%333DDD

; Semi-universal text cursor

@cursor		DT 	%4AD

@kbuf_length	DT 	$0
@keybuffer	REST 	$40
@kb_null	DT 	$0
@kb_got_line	DT 	$0

; -- Get string --
; 
;

@getstring
		LDA 	@kb_got_line		; Wait for string
		JEQ 	@getstring

		SEI				; Disable interrupts 
						;(no typing while
						; loading the buffer)

		LDA 	#$0			; Clear got_line flag
		STA 	@kb_got_line

		LDX 	#$0
@.copy
		LDA 	@keybuffer,X		; Copy the interrupt-buffer 
						; to the gets
		STA	@.buffer,X		;  buffer
		INX
		TXA
		CMP 	#$40
		JNE 	@.copy

		CLI				; Restore interrupts

		RST

@.buffer	REST	$40			; Character buffer
		DT 	$0

;  ----
;   Feed screen down one line
;   Arguments: None
;
;   Not pretty, but does the work
;  ----


@feedscreen
		PSH	A			; Save A

		LDA 	#%4AD
		STA 	@cursor			; Restore cursor

		LDA 	#%DDB			; Set from and to- page to %DDB
		STA 	@.to
		STA 	@.from
		LDA 	#%DDD		
		STA 	@.to+$1			; Set to index
		LDA 	#%D2D	
		STA 	@.from+$1		; Set from index

		SEI				; Mask interrupts
						; This will stop redrawing
@.loop
		LDA 	(@.from)		; Transfer from -> to
		STA 	(@.to)

		LDA 	@.from			; Increase from pointer by 1
		INC 	@.from+$1
		ADD 	#%000
		STA 	@.from

		LDA 	@.to			; Increase to pointer by 1
		INC 	@.to+$1
		ADD 	#%000
		STA 	@.to

		LDA 	@.to			; Loop if not on end-page
		CMP 	#%DD0
		JNE 	@.loop

		CLI				; Resume interrupts
		INC 	@redraw			; Redraw screen

		PLL	A			; Restore A
		RST				; Return

@.to		DT 	%DDB %DDD
@.from		DT 	%DDB %D2D


; ----------------------------------
; Put character
; Arguments: Character (A)
;
; Uses @cursor
; ----------------------------------

@putchar
		LDX	@cursor
		STA	@screen.page2,X

		INX

		JVC	+$5
		LDX	#%4AD
		JSR	@feedscreen

		STX	@cursor

		INC 	@redraw			; Redraw screen

		RST

; ----------------------------------
;  Put number in nonary
;  Arguments: Number (A)
; ----------------------------------
@putnon
		PSH	A			; Save A

		AND 	#%400
		LSR
		LSR
		LSR
		LSR
		TAX
		LDA 	@.nonc+$4,X		; Compensate for the possible
						; negative index of -4
		JSR	@putchar

		PLL	A			; Restore and re-save A
		PSH	A

		AND 	#%040
		LSR
		LSR
		TAX
		LDA 	@.nonc+$4,X
		JSR	@putchar

		PLL	A			; Restore A

		AND 	#%004
		TAX
		LDA 	@.nonc+$4,X
		JSR	@putchar

		RST

@.nonc		DT	'DCBA01234'

; ----------------------------------
;  Put string
;  Arguments: String page (X), String offset (Y)
;  String is 0-ended
; ----------------------------------

@puts
		SEI				; Disable interrupts
		PSH	A

		STX 	@.string		; Store string in memory
		STY 	@.string+$1	
@.loop
		LDA 	(@.string)	
		JEQ 	@.done			; Continue until null

		JSR	@putchar

		LDA 	@.string
		INC 	@.string+$1
		ADD 	#$0			; Add with carry
		STA 	@.string

		JMP 	@.loop			;
@.done
		PLL	A			; Restore accumulator
		CLI				; Restore interrupts
		INC 	@redraw			; Redraw screen
		RST				; Return

@.string	DW 	$0

; -------------------------------------
; String comparison
; Arguments:
;   (pushed in this order)
;   String1 page (stack), String1 offset (stack), 
;   String2 page (stack), String2 offset (stack)
; Returns
;   0 if equal (in acc)
; -------------------------------------


@strcmp
		PLX				; Save return address
		PLY	

		; Get string addresses
		PLL	@.str2+$1
		PLL 	@.str2+$0
		PLL	@.str1+$1
		PLL	@.str1+$0

		PHY				; Restore return address
		PHX

@.loop
		LDA 	(@.str1)
		CMP 	(@.str2)
		JNE 	@.mismatch

		LDA 	(@.str1)
		JEQ 	@.match
	
		LDA 	@.str1
		INC 	@.str1+$1
		ADD 	#$0
		STA 	@.str1
	
		LDA 	@.str2
		INC 	@.str2+$1
		ADD 	#$0
		STA 	@.str2

		JMP 	@.loop
		
@.mismatch
		PLX				; Save return address
		PLY
	
		LDA 	#$1			; Return 1
	
		PHY				; Restore return address
		PHX
	
		RST				; Return
@.match
		PLX				; Save return address
		PLY
	
		LDA 	#$0			; Return 0
	
		PHY				; Restore return address
		PHX
	
		RST				; Return

@.str1 		DW 	$1
@.str2 		DW 	$1
	
; ------------------------------------
;  String length
;  Arguments: string page (X), string offset (Y)
;  Returns: string length (A)
; ------------------------------------

@strlen	
		STX	@.str
		STY	@.str+$1
		LDX	#$0

@.loop
		LDA	(@.str)
		JEQ	@.done

		LDA	@.str
		INC	@.str+$1
		ADD	#$0
		STA	@.str

		INX

		JMP	@.loop
@.done
		TXA
		RST


@.str		DW 	$0

; -- index, ... like stdlib:s index() --
; Arguments: A -- Character to match,
;            X, Y -- Pointer to string
; Returns:
;	     X,Y -- Pointer to position in string matching A
;			or 0,0 if zero

@index
@.loop
		CMP	 X,Y			; Compare to given character
		JEQ 	@.match			;

		PSH	A
		LDA 	X,Y			; Compare to zero
		JEQ 	@.mmatch
		PLL	A

		PSH	A			; Preserve A
		TXA				; Increase pointer by one
		INY
		ADD 	#$0
		TAX
		PLL	A

		JMP 	@.loop
@.match
		RST
@.mmatch
		PLL	A
		LAD 	@null
		RST
; -- strspn (works like stdlib's epinymous function) --
; Arguments: 
;    Pushed in this order: 
;    (1) String page
;    (2) String offset
;    (3) Valid characters page
;    (4) Valid characters offset
; Returns
;   A: Length of span in primary string consisting 
;      entirely of characters in second string
;

@strspn
		PLX				; Save return address
		PLY	

		PLL	@.valid+$1
		PLL	@.valid
		PLL	@.string+$1
		PLL	@.string

		PHY
		PHX

		LDA	#$0
		STA	@.length

@.loop
		; Check if string is null
		LDA	(@.string)
		JEQ	@.done

		; Search for current character in valid characters
		LAD	@.valid
		JSR	@index
		TXA
		JNE	@.done
		TYA
		JNE	@.done

		; Increase span
		INC	@.length

		; Increase string pointer
		;
		LDA	@.string
		INC	@.string+$1
		ADD	#$0
		STA	@.string

		JMP	@.loop
@.done
		LDA 	@.length
		RST

@.string 	DW 	$0
@.valid  	DW 	$0
@.length 	DT 	$0
; -- strcspn (works like stdlib's epinymous function) --
; Arguments: 
;    Pushed in this order: 
;    (1) String page
;    (2) String offset
;    (3) Invalid characters page
;    (4) Invalid characters offset
; Returns
;   A: Length of span in primary string consisting 
;      entirely of characters in second string
;

@strcspn
		PLX			; Save return address
		PLY	

		PLL	@.invalid+$1
		PLL	@.invalid
		PLL	@.string+$1
		PLL	@.string

		PHY
		PHX

		LDA 	#$0
		STA 	@.length

@.loop
		; Check if string is null
		LDA 	(@.string)
		JEQ 	@.done

		; Search for current character in invalid characters
		LAD 	@.invalid
		JSR 	@index
		TXA
		JNE 	+$4
		TYA
		JEQ 	@.done

		; Increase span
		INC 	@.length

		; Increase string pointer
		;
		LDA 	@.string
		INC 	@.string+$1
		ADD 	#$0
		STA 	@.string
@.done
		LDA 	@.length
		RST

@.string 	DW 	$0
@.invalid  	DW 	$0
@.length 	DT 	$0

; -- memset (6 trit version) --
; Arguments, pushed in this order: Value, Length, Memory page, Memory offset
;
;

@memset6
		PLX
		PLY

		PLL	@.memory+$1
		PLL	@.memory
		PLL	@.length
		PLL	@.character

		PHY
		PHX

@.loop
		LDA 	@.length
		LAD	@.memory
		DEBUG
		LDA	@.character
		STA	(@.memory)

		LDA 	@.memory
		INC	@.memory+$1
		ADD	#$0
		STA	@.memory

		DEC	@.length
		JNE	@.loop

		RST
		
@.length	DT	$0	
@.memory	DW	$0
@.character	DT	$0

